#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdarg.h>
#include <assert.h>
#include <GL/glut.h>

#define NSEG	12

void disp(void);
void reshape(int x, int y);
void keyb(unsigned char key, int x, int y);
void glut_printf(int x, int y, const char *fmt, ...);


static int win_width, win_height;
static int single;
static int z0, preclip, have_preclip;


int main(int argc, char **argv)
{
	char *ext;

	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
	glutInitWindowSize(800, 600);
	glutCreateWindow("VPro Z=0 bug test");

	glutDisplayFunc(disp);
	glutIdleFunc(glutPostRedisplay);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyb);

	glEnable(GL_CULL_FACE);

	ext = (char*)glGetString(GL_EXTENSIONS);
	have_preclip = strstr(ext, "SGIX_vertex_preclip") ? 1 : 0;

	glutMainLoop();
	return 0;
}


void disp(void)
{
	int i;
	float x, y, z, xp, yp, theta, t;
	static const unsigned char colors[][4] = {
		{0xff, 0, 0, 0xff},
		{0xff, 0x80, 0x40, 0xff},
		{0, 0xff, 0, 0xff},
		{0x40, 0x80, 0xff, 0xff},
		{0xff, 0, 0xff, 0xff},
		{0x40, 0xff, 0x80, 0xff},
		{0xff, 0xff, 0, 0xff},
		{0xff, 0x40, 0xff, 0xff},
		{0, 0xff, 0xff, 0xff},
		{0xff, 0xff, 0x40, 0xff},
		{0, 0, 0xff, 0xff},
		{0x40, 0xff, 0xff, 0xff}
	};

	t = (float)glutGet(GLUT_ELAPSED_TIME) / 800.0f;

	glClearColor(0.1, 0.1, 0.1, 1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	/* draw tunnel */
	z = z0 ? 0.0f : -0.01f;

	glBegin(GL_QUADS);
	xp = cos(t);
	yp = sin(t);
	for(i=0; i<NSEG; i++) {
		theta = ((float)(i + 1) / (float)NSEG) * (M_PI * 2.0);
		x = cos(theta + t);
		y = sin(theta + t);
		glColor4ubv(colors[i]);
		glVertex3f(xp, yp, z);
		glVertex3f(x, y, z);
		glVertex3f(x, y, -50.0);
		glVertex3f(xp, yp, -50.0);
		xp = x;
		yp = y;

		if(single) break;
	}
	glEnd();


	/* draw text window */
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, win_width, win_height, 0, -1, 1);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glColor4f(0, 0, 0, 0.65f);
	glRecti(0, 55, win_width, 0);
	glDisable(GL_BLEND);
	glPopMatrix();

	glColor3f(0, 0, 0);
	glut_printf(12, 22, "start Z = %f (press 'z' to toggle)", z);
	glColor3f(0, 1, 0);
	glut_printf(10, 20, "start Z = %f (press 'z' to toggle)", z);

	glColor3f(0, 0, 0);
	glut_printf(12, 42, "vertex pre-clip: %s (%s)", preclip ? "on" : "off",
			have_preclip ? "press 'c' to toggle" : "extension unavailable");
	glColor3f(0, 1, 0);
	glut_printf(10, 40, "vertex pre-clip: %s (%s)", preclip ? "on" : "off",
			have_preclip ? "press 'c' to toggle" : "extension unavailable");


	glutSwapBuffers();
	assert(glGetError() == GL_NO_ERROR);
}

void reshape(int x, int y)
{
	win_width = x;
	win_height = y;

	glViewport(0, 0, x, y);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(50.0, (float)x / (float)y, 0.5, 500.0);
}

void keyb(unsigned char key, int x, int y)
{
	switch(key) {
	case 27:
		exit(0);

	case 's':
		single ^= 1;
		break;

	case 'z':
		z0 ^= 1;
		break;

	case 'c':
		if(have_preclip) {
			preclip ^= 1;
			if(preclip) {
				glEnable(GL_VERTEX_PRECLIP_SGIX);
				glHint(GL_VERTEX_PRECLIP_HINT_SGIX, GL_FASTEST);
			} else {
				glDisable(GL_VERTEX_PRECLIP_SGIX);
			}
		}
		break;

	default:
		break;
	}
}

void glut_printf(int x, int y, const char *fmt, ...)
{
	va_list ap;
	char *p;
	char buf[512];

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, win_width, win_height, 0, -1, 1);

	glRasterPos2i(x, y);

	va_start(ap, fmt);
	vsnprintf(buf, sizeof buf, fmt, ap);
	va_end(ap);

	p = buf;
	while(*p) {
		glutBitmapCharacter(GLUT_BITMAP_9_BY_15, *p++);
	}

	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}
